<?php

namespace Drupal\tag_formatter\Plugin\Field\FieldFormatter;

use Drupal\Core\Field\Plugin\Field\FieldFormatter\EntityReferenceLabelFormatter;
use Drupal\Core\Entity\Exception\UndefinedLinkTemplateException;
use Drupal\Core\Field\FieldItemListInterface;
use Drupal\Core\Form\FormStateInterface;
//use Drupal\taxonomy\Entity\Term;

/**
 * Plugin implementation of the 'entity_reference_tag_formatter' formatter.
 * 
 * @FieldFormatter(
 *  id = "entity_reference_tag_formatter",
 *  label = @Translation("Tag Formatter"),
 *  description = @Translation("Display the entity label as tag."),
 *  field_types = {
 *   "entity_reference"
 *  }
 * )
 */
class TagFormatter extends EntityReferenceLabelFormatter {

  /**
   * {@inheritdoc}
   */
  public static function defaultSettings() {
    return [
        'link_absolute' => FALSE,
        'letter_mod' => 'none',
        'non_letter_mod' => 'none',
        'trim' => FALSE,
        'transliterate' => FALSE,
        'add_hash' => FALSE,
        'use_term_remap' => FALSE,
        'remap_enabled' => FALSE,
        'remap_rules' => '|',
        'tag_separator' => '',
        'tag_format' => 'field',
        ] + parent::defaultSettings();
  }

  /**
   * {@inheritdoc}
   */
  public function settingsForm(array $form, FormStateInterface $form_state) {
    $element = parent::settingsForm($form, $form_state);
    $element['link_absolute'] = [
        '#type'          => 'checkbox',
        '#title'         => $this->t('Absolute link'),
        '#description'   => $this->t('Use absolute URL for link.'),
        '#default_value' => !empty($this->getSetting('link_absolute')),
        ];
    $element['letter_mod'] = [
        '#type'          => 'select',
        '#title'         => $this->t('Letter case'),
        '#description'   => $this->t('Letters transformation.'),
        '#default_value' => $this->getSetting('letter_mod'),
        '#options'       => [
            'none'   => $this->t('None'),
            'upper'  => $this->t('TO UPPERCASE'),
            'lower'  => $this->t('to lowercase'),
            'camel'  => $this->t('To Camel Case'),
            ],
        ];
    $element['non_letter_mod'] = [
        '#type'          => 'select',
        '#title'         => $this->t('Non-Letters'),
        '#description'   => $this->t('Non-Letters replacement.'),
        '#default_value' => $this->getSetting('non_letter_mod'),
        '#options'       => [
            'none'   => $this->t('None'),
            'remove' => $this->t('Remove'),            
            '_'      => $this->t('_ (underscore)'),
            '-'      => $this->t('- (dash)'),
            ' '      => $this->t('Space'),
            ],
        ];
    $element['trim'] = [
        '#type'          => 'checkbox',
        '#title'         => $this->t('Trim'),
        '#description'   => $this->t('Trim label.'),
        '#default_value' => !empty($this->getSetting('trim')),
        ];        
    $element['transliterate'] = [
        '#type'          => 'checkbox',
        '#title'         => $this->t('Transliterate'),
        '#description'   => $this->t('Transliterate label.'),
        '#default_value' => !empty($this->getSetting('transliterate')),
        ];
    $element['add_hash'] = [
        '#type'          => 'checkbox',
        '#title'         => $this->t('Hash symbol'),
        '#description'   => $this->t('Prepend label with # symbol.'),
        '#default_value' => !empty($this->getSetting('add_hash')),
        ];
    $element['remap_enabled'] = [
        '#type'          => 'checkbox',
        '#title'         => $this->t('Remap enabled'),
        '#description'   => $this->t('Enable or disable remap rules.'),
        '#default_value' => !empty($this->getSetting('remap_enabled')),
        ];
    $element['remap_rules'] = [
        '#type'          => 'textarea',
        '#title'         => $this->t('Remap rules'),
        '#description'   => $this->t('One rule per line: "term in set 1|term in set 2 | prefix | suffix"<br>Omit source term to setup default values.<br>All terms must be pre-formatted, except # adding.<br>Note: prefix and suffix can contain some html wrappers.'),
        '#default_value' => $this->getSetting('remap_rules'),
        '#placeholder'   => "Term_A | Term_A_in_set2 \nTerm_B || <b> | </b>\nTerm_C | Term_C_in_set2\n| Term_Default_in_set2 | <i> | </i>",
        ];
    $element['tag_separator'] = [
        '#type'          => 'textfield',
        '#title'         => $this->t('Separator'),
        '#description'   => $this->t('The separator to use, include spaces. Can contain some html.'),
        '#default_value' => $this->getSetting('tag_separator'),
        '#size' => 10,
        ];
    $element['tag_format'] = [
        '#type'          => 'select',
        '#title'         => $this->t('Tag format'),
        '#description'   => $this->t('Use default field formatter or merge multiple values in one item.'),
        '#default_value' => $this->getSetting('tag_format'),
        '#options'       => [
            'field' => $this->t('Field'),
            'merged_field' => $this->t('Merged field'),
            ],
        ];
    return $element;
  }

  /**
   * {@inheritdoc}
   */
  public function settingsSummary() {
    $summary = parent::settingsSummary();
    if (!empty($this->getSetting('link_absolute'))) $summary[] = $this->t('Absolute link');
    $s = t('Letter mod: "@letter"', ['@letter' => $this->getSetting('letter_mod')]);
    $s.= ', ' . t('Non-letter mod: "@nonletter"', ['@nonletter' => $this->getSetting('non_letter_mod')]);
    if (!empty($this->getSetting('trim'))) $s .', ' . t('Trim');
    if (!empty($this->getSetting('transliterate'))) $s .= ', ' . t('Transliterate');
    if (!empty($this->getSetting('remap_enabled'))) $s .= ', ' . t('Remap');
    if (!empty($this->getSetting('add_hash'))) $s .= ', ' . t('Add hash');
    $s .= ', ' . t('Separator: "@separator"',  ['@separator' => $this->getSetting('tag_separator')]);
    $s .= ', ' . t('Format: "@format"',  ['@format' => $this->getSetting('tag_format')]);
    $summary[] = $s;
    return $summary;
  }

  /**
   * {@inheritdoc}
   */
  public function viewElements(FieldItemListInterface $items, $langcode) {
    // Replacement for EntityReferenceLabelFormatter::viewElements()
    $elements = [];
    $output_as_link = !empty($this->getSetting('link'));
    $link_absolute = !empty($this->getSetting('link_absolute')); 
    
    // Get label formatting parameters
    $letter_mod = $this->getSetting('letter_mod');
    if (!empty($letter_mod)) {
        $mods = [ 
            'camel' => MB_CASE_TITLE,
            'lower' => MB_CASE_LOWER,
            'upper' => MB_CASE_UPPER,
            ];
        if (array_key_exists($letter_mod, $mods)) $letter_mod = $mods[$letter_mod];
        else unset($letter_mod);
    }
    $non_letter_mod = $this->getSetting('non_letter_mod');
    if (empty($non_letter_mod)) $non_letter_mod = 'none';
    else if ($non_letter_mod == 'remove') $non_letter_mod = '';
    $trim = !empty($this->getSetting('trim'));
    $transliterate = !empty($this->getSetting('transliterate'));
    $add_hash = !empty($this->getSetting('add_hash'));

    // Parse remap rules
    $map = [];
    $map_default = [];
    $map_enabled = !empty($this->getSetting('remap_enabled'));
    if ($map_enabled) {
        $remap = $this->getSetting('remap_rules');
        if (!empty($remap)) {
            $lines = explode("\n", $remap);
            foreach ($lines as $s) {
                $s = trim($s);
                if (empty($s)) continue;
                $m = explode("|", $s, 4);
                $m0 = trim($m[0]);
                if (empty($m0)) {
                    // default rule
                    $map_default['label']  = (isset($m[1]))? trim($m[1]) : '';
                    $map_default['prefix'] = (isset($m[2]))? trim($m[2]) : '';
                    $map_default['suffix'] = (isset($m[3]))? trim($m[3]) : '';
                }
                else {
                    $m1 = (isset($m[1]))? trim($m[1]) : $m0;
                    $map[$m0]['label']  = (empty($m1))? $m0 : $m1;
                    $map[$m0]['prefix'] = (isset($m[2]))? trim($m[2]) : '';
                    $map[$m0]['suffix'] = (isset($m[3]))? trim($m[3]) : '';
                }
            }
        }
    }

    foreach ($this->getEntitiesToView($items, $langcode) as $delta => $entity) {
        $label = $entity->label();
        $suffix = $prefix = '';
     
        // Format label
        if ($non_letter_mod != 'none') {
            $label = preg_replace('/\P{L}+/u',' ',$label);
        }
        if ($trim) $label = trim($label);
        if ($transliterate) $label = \Drupal::transliteration()->transliterate($label);
        if (!empty($letter_mod)) $label = mb_convert_case($label, $letter_mod, "UTF-8");
        if (($non_letter_mod != 'none') && ($non_letter_mod != ' ')) {
            $label = str_replace(' ', $non_letter_mod, $label);
        }

        if ($map_enabled) {
            if (array_key_exists($label, $map)) {
                $prefix = $map[$label]['prefix'];
                $suffix = $map[$label]['suffix'];
                $label  = $map[$label]['label'];
            }
            else if (!empty($map_default)) {
                $prefix = $map_default['prefix'];
                $suffix = $map_default['suffix'];
                if (!empty($map_default['label'])) $label = $map_default['label'];
            }
            else continue;
        }
        if ($add_hash) $label = '#'.$label;
     
        // If the link is to be displayed and the entity has a uri, display a
        // link.
        if ($output_as_link && !$entity->isNew()) {
            try {
                $url = $entity->toUrl();
                
            }
            catch (UndefinedLinkTemplateException $e) {
        // This exception is thrown by \Drupal\Core\Entity\Entity::toUrl()
        // and it means that the entity type doesn't have a link template nor
        // a valid "uri_callback", so don't bother trying to output a link for
        // the rest of the referenced entities.
                $output_as_link = FALSE;
            }
        }

        if ($output_as_link && isset($url) && !$entity->isNew()) {
            $url->setAbsolute($link_absolute);
            $elements[$delta] = [
                '#type'    => 'link',
                '#title'   => $label,
                '#url'     => $url,
                '#options' => [],
            //    '#options' => $url->getOptions(),
            ];

            if (!empty($items[$delta]->_attributes)) {
                $elements[$delta]['#options'] += ['attributes' => []];
                $elements[$delta]['#options']['attributes'] 
                    += $items[$delta]->_attributes;
        // Unset field item attributes since they have been included in the
        // formatter output and shouldn't be rendered in the field template.
                unset($items[$delta]->_attributes);
            }
        }
        else {
            $elements[$delta] = ['#plain_text' => $label];
        }
        $elements[$delta]['#cache']['tags'] = $entity->getCacheTags();
        if (!empty($prefix)) $elements[$delta]['#prefix'] = $prefix;
        if (!empty($suffix)) $elements[$delta]['#suffix'] = $suffix;
    }

    // Insert separators
    $tag_separator = $this->getSetting('tag_separator');
    if (!empty($tag_separator)) {
        $prev = NULL;
        foreach ($elements as $delta => $row) {
            if (isset($prev)) {
                if (isset($elements[$prev]['#suffix']))
                    $elements[$prev]['#suffix'] .= $tag_separator;    
                else
                    $elements[$prev]['#suffix'] = $tag_separator;
            }
            $prev = $delta;
        }
    }

    // Merge multiple values, if needed.
    $tag_format = $this->getSetting('tag_format');
    if (isset($tag_format) && ($tag_format == 'merged_field')) {
        $out = '';
        $ctags = [];
        $renderer = \Drupal::service('renderer');
        foreach ($elements as $delta => $row) {
            $ctags = array_merge($ctags, $row['#cache']['tags']);
            $out .= $renderer->render($row);
        }
        $printed = [];
        $printed[0] = [
            '#children' => $out,
            '#cache' => ['tags' => $ctags],
            ];
        return $printed;
    }

    return $elements;
  }

}
